from collections import defaultdict

import holoviews as hv
import numpy as np
import pandas as pd

from rnaseq_lib.plotting import plot_boxplot
from rnaseq_lib.tissues import get_gtex_samples, return_samples
from rnaseq_lib.tissues import get_normal_samples
from rnaseq_lib.tissues import get_tumor_samples


def gene_expression_boxplot(df, tissue, gene, normal=False):
    """
    Returns a holoviews box and whisker object of a gene for a tissue

    :param pd.DataFrame df: Gene by Samples DataFrame
    :param str tissue: Tissue to use
    :param str gene: Gene to use
    :param bool normal: If True, will include TCGA normal samples
    :return: Boxplot of gene for tissue
    :rtype: hv.BoxWhisker
    """
    plot_info = {'GTEx': get_gtex_samples(tissue),
                 'Tumor': get_tumor_samples(tissue)}

    if normal:
        plot_info['Normal'] = get_normal_samples(tissue)

    return plot_boxplot(df, plot_info=plot_info, feature=gene, title=tissue, norm_func=lambda x: np.log2(x + 1))


def get_tcga_gtex_label_info():
    """
    :return: Returns TCGA and GTEx label, tissue, and samples information
    :rtype: defaultdict
    """
    samples = return_samples()
    info = defaultdict(list)
    for tissue in sorted(samples):
        info['label'].extend(['Tumor' if x.endswith('-01')
                              else 'Normal' if x.endswith('-11') else 'GTEx' for x in samples[tissue]])
        info['sample'].extend([x for x in samples[tissue]])
        info['tissue'].extend([tissue for _ in samples[tissue]])
    return info



# -*- coding: utf-8 -*-
"""
Created on Sat May 27 12:46:25 2017

@author: ehsanamid
"""

# -*- coding: utf-8 -*-
"""
Created on Fri May 26 12:17:14 2017

@author: ehsanamid
"""

from sklearn.neighbors import NearestNeighbors as knn
#from sklearn.decomposition import PCA
from sklearn.decomposition import TruncatedSVD
import numpy as np
import pylab as plot
    
def generate_triplets(X, k=20):
    num_extra = np.maximum(k+10, 30) # look up more neighbors
    n = X.shape[0]
    nbrs = knn(n_neighbors= num_extra + 1, algorithm='auto').fit(X)
    distances, indices = nbrs.kneighbors(X)
    #    sig = distances[:,10]
    sig = np.mean(distances[:, 10:15], axis=1) # scale parameter
    P = np.exp(-distances**2/np.reshape(sig[indices.flatten()],[n, num_extra + 1])/sig[:,np.newaxis])
    sort_indices = np.argsort(-P, axis = 1) # actual neighbors
    triplets = np.zeros([n * k**2, 3])
    weights = np.zeros(n * k**2)
    cnt = 0
    for i in xrange(n):
        for j in xrange(k):
            sim = indices[i,sort_indices[i,j+1]]
            p_sim = P[i,sort_indices[i,j+1]]
            rem = indices[i,sort_indices[i,:j+2]].tolist()
            l = 0
            while (l < k):
                out = np.random.choice(n)
                if out not in rem:
                    triplets[cnt,:] = [i, sim, out]
                    p_out = np.exp(-np.sum((X[i,:]-X[out,:])**2)/(sig[i] * sig[out]))
                    if p_out < 1e-20:
                        p_out = 1e-20
                    weights[cnt] = p_sim/p_out
                    rem.append(out)
                    l += 1
                    cnt += 1
        if ((i+1) % 500) == 0:
            pass
            #print 'Genareted triplets %d out of %d' % (i+1, n)
    weights /= np.max(weights)
    weights += 0.01
    weights = np.log(1 + 50 * weights)
    weights /= np.max(weights)
    return (triplets.astype(int), weights.flatten())
        
def tete_grad(Y, triplets, weights):
    n, dim = Y.shape
    grad = np.zeros([n, dim])
    y_ij = Y[triplets[:,0],:] - Y[triplets[:,1],:]
    y_ik = Y[triplets[:,0],:] - Y[triplets[:,2],:]
    d_ij = 1 + np.sum(y_ij**2,axis=-1)
    d_ik = 1 + np.sum(y_ik**2,axis=-1)
    num_viol = np.sum(d_ij > d_ik)
    denom = (d_ij + d_ik)**2
    loss = weights.dot(d_ij/(d_ij + d_ik))
    gs = 2 * y_ij * (d_ik/denom * weights)[:,np.newaxis]
    go = 2 * y_ik * (d_ij/denom * weights)[:,np.newaxis]
    for i in range(dim):
        grad[:,i] += np.bincount(triplets[:,0], weights= gs[:,i] - go[:,i])
        grad[:,i] += np.bincount(triplets[:,1], weights = -gs[:,i])
        grad[:,i] += np.bincount(triplets[:,2], weights = go[:,i])
    return (loss, grad, num_viol)
    

def tete(X, num_dims = 2, num_neighbs = 20):
    n, dim = X.shape
    X -= np.min(X)
    X /= np.max(X)
    X -= np.mean(X,axis=0)
    if dim > 50:
#        pca = PCA(n_components=50)
#        pca.fit(X)
#        X = np.dot(X, pca.components_.transpose())
        X = TruncatedSVD(n_components=50, random_state=0).fit_transform(X)
    Y = np.random.normal(size=[n, 2]) * 0.0001
    C = np.inf
    best_C = np.inf
    best_Y = Y
    tol = 1e-7
    num_iter = 2000
    eta = 500.0 # learning rate
    
    triplets, weights = generate_triplets(X, num_neighbs)
    num_triplets = float(triplets.shape[0])
    
    for itr in range(num_iter):
        old_C = C
        C, grad, num_viol = tete_grad(Y, triplets, weights)
        
        # maintain best answer
        if C < best_C:
            best_C = C
            best_Y = Y
            
        # update Y
        Y -= (eta/num_triplets * n) * grad;
        
        # update the learning rate
        if old_C > C + tol:
            eta = eta * 1.01
        else:
            eta = eta * 0.5
        
        if (itr+1) % 100 == 0:
            pass
            # print 'Iteration: %4d, Loss: %3.3f, Num viol: %0.4f' % (itr+1, C, float(num_viol)/num_triplets)
    #np.savetxt('result_tete_outlier_10000.txt',best_Y)
    return best_Y


if __name__ == "__main__":
    print "Run Y = tete.tete(X, num_dims, num_neighbs) to perform t-ETE on your dataset."
    X = np.loadtxt("mnist2500_X.txt")
    labels = np.loadtxt("mnist2500_labels.txt")
    Y = tete(X, 2, 20)
    plot.scatter(Y[:,0], Y[:,1], 10, labels)
    plot.show();




import os
from subprocess import call

import logging
logging.basicConfig(level=logging.INFO)


def filter_aligned_reads(bam_path, output_name=None, paired=True):
    """
    Filters bams for aligned reads

    :param str bam_path: Path to bam file
    :param str output_name: Defaults to input bam with '.filtered' inserted before the .bam extension
    :param bool paired: Only keep paired reads
    :return: Path to filtered bam
    :rtype: str
    """
    if not output_name:
        output_name = os.path.basename(os.path.splitext(bam_path)[0]) + '.filtered.bam'

    # Define parameters
    work_dir = os.path.dirname(os.path.abspath(bam_path))
    parameters = ['docker', 'run',
                  '-v', '{}:/data'.format(work_dir),
                  'quay.io/ucsc_cgl/samtools',
                  'view',
                  '-h',
                  '-o', '/data/{}'.format(output_name),
                  '-F', '0x04 ']
    if paired:
        parameters.extend(['-f', '0x02'])
    parameters.append(os.path.join('/data', os.path.basename(bam_path)))

    # Call tool
    output_path = os.path.join(work_dir, output_name)
    if not os.path.exists(output_path):
        logging.info('Filtering bam')
        call(parameters)
    else:
        logging.info('Skipping. Filtered bam exists: {}'.format(output_path))
    return output_path


def make_test_bam(bam_path, region='chr6', output_name='chr6.test.bam', clean_input_bam=False):
    """
    Makes a smaller bam based on the region argument passed

    :param str bam_path: Path to bam to use to make test bam
    :param str region: Region of the genome to subset from
    :param str output_name: Output file name test bam
    :param bool clean_input_bam: Cleans the input bam before starting
    :return: Path to test bam
    :rtype: str
    """
    if clean_input_bam:
        bam_path = filter_aligned_reads(bam_path)

    work_dir = os.path.dirname(os.path.abspath(bam_path))
    docker_parameters = ['docker', 'run',
                         '-v', '{}:/data'.format(work_dir),
                         'quay.io/ucsc_cgl/samtools']

    bam_no_ext = bam_path.split('.bam')[0]
    if not os.path.exists(bam_no_ext + '.bai') and not os.path.exists(bam_no_ext + '.bam.bai'):
        index = docker_parameters + ['index', os.path.join('/data', os.path.basename(bam_path))]
        call(index)

    parameters = docker_parameters + ['view',
                                      '-b',
                                      '-h',
                                      '-o', os.path.join('/data', output_name),
                                      os.path.join('/data', os.path.basename(bam_path)),
                                      region]
    call(parameters)
    return os.path.join(work_dir, output_name)  #filter_aligned_reads()



import pandas as pd
import progressbar
import requests

bar = progressbar.ProgressBar()


def create_civic_drug_disease_dataframe():
    """
    Hits CIViC's API to build a dataframe on genes, cancers, and drugs

    :return: Dataframe of CIVIC information
    :rtype: pd.DataFrame
    """
    payload = {'count': 999}  # Only 295 genes at time of writing
    request = requests.get('https://civic.genome.wustl.edu/api/genes/', params=payload)
    assert request.status_code == 200, 'Request failed: {}'.format(request.status_code)

    records = request.json()['records']

    # Create records in the format:  Cancer, Gene, Drugs, Variant-Name, Aliases, Description
    info = []
    for record in bar(records):
        gene = record['name']
        description = record['description']
        aliases = ','.join(record['aliases'])
        variants = record['variants']

        for variant in variants:
            var_name = variant['name']
            var_id = variant['id']
            var = requests.get('https://civic.genome.wustl.edu/api/variants/{}'.format(var_id)).json()
            for evidence in var['evidence_items']:
                disease = evidence['disease']['name']
                drugs = ','.join([x['name'] for x in evidence['drugs']])
                info.append((disease, drugs, gene, aliases, var_name, description))

    # Create DataFrame
    labels = ['Cancer', 'Drugs', 'Gene', 'Aliases', 'Variant-Name', 'Description']
    return pd.DataFrame.from_records(info, columns=labels)



import os
import pickle

import pandas as pd

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


def return_samples():
    """
    Returns sample dictionary which maps TCGA and GTEx samples to a tissue.
    Synapse ID: syn10296681

    :return: Tissues are keys are list of samples are values
    :rtype: dict(str, list(str))
    """
    return pickle.load(open(os.path.join(os.path.dirname(__location__), 'utils/data/samples.pickle'), 'rb'))


def get_gene_map():
    """
    Dictionary mapping gene ID to gene name

    :return: Gene map
    :rtype: dict
    """
    return pickle.load(open(os.path.join(os.path.dirname(__location__), 'utils/data/gene_map.pickle'), 'rb'))


def map_genes(df):
    gene_map = get_gene_map()
    genes = [gene_map[x.split('.')[0]] for x in df.index]
    df.index = genes
    return df


def get_mab_targets():
    """
    Returns sorted list of MAB cancer drug targets

    :return: Sorted gene list
    :rtype: list
    """
    path = os.path.join(__location__, 'data/cancer-MAB-gene-targets.txt')
    return sorted([x.strip() for x in open(path, 'r').readlines()])


def get_ucsf_genes():
    """
    Returns sorted list of UCSF genes

    :return: Sorted gene list
    :rtype: list
    """
    path = os.path.join(__location__, 'data/UCSF-genes.csv')
    return sorted([x.strip() for x in open(path, 'r').readlines()])


def get_civic_genes():
    """
    Returns sorted list of genes from CIViC

    :return: Sorted gene list
    :rtype: list
    """
    path = os.path.join(__location__, 'data/civic-genes.txt')
    return sorted([x.strip() for x in open(path, 'r').readlines()])


def get_ucsf_subset(df):
    """
    Subset UCSF dataframe and return.

    :param pd.DataFrame df: Input Dataframe in the format of "Genes by Samples"
    :return: Subset of Dataframe that only includes UCSF genes
    :rtype: pd.DataFrame
    """
    df = map_genes(df)

    ucsf_genes = get_ucsf_genes()
    ucsf_genes = [x for x in ucsf_genes if x in df.index]

    return df.loc[ucsf_genes]


def get_tumor_samples(tissue):
    """
    Returns TCGA tumor samples for a tissue
    :param str tissue:
    :return: List of tumor samples
    :rtype: list
    """
    samples = return_samples()
    return [x for x in samples[tissue] if x.endswith('-01')]


def get_gtex_samples(tissue):
    """
    Returns GTEx samples for a tissue

    :param str tissue:
    :return: List of GTEx samples
    :rtype: list
    """
    samples = return_samples()
    return [x for x in samples[tissue] if not x.startswith('TCGA')]


def get_normal_samples(tissue):
    """
    Returns TCGA normal samples for a tissue
    :param str tissue:
    :return: List of TCGA normal samples
    :rtype: list
    """
    samples = return_samples()
    return [x for x in samples[tissue] if x.endswith('-11')]
